(* Author: Vincent Cremet *)
(* Date: July 2006 *) 
(* File: MatchCompNat.v *)
(* Description: Matching operators for comparing integers *)

(** * Matching Integer Comparisons *)

Require Import Arith.
Require Import Omega.

(*****************************************************************)

Lemma Match2:
  forall (A B: Prop),
    { A } + { B } ->
    forall (S: Set) (xA xB: S),
      { x: S | (A /\ x = xA) \/ (B /\ x = xB) }
.

Proof.
intros.
induction H.
exists xA. tauto.
exists xB. tauto.
Qed.

Lemma Match2_diff:
  forall (A B: Prop),
    { A } + { B } ->
    (A -> ~B) ->
    (B -> ~A) ->
    forall (S: Set) (xA xB: S),
      { x: S |
        (A -> (x = xA)) /\
        (B -> (x = xB))
      }
.

Proof.
intros.
induction H.
exists xA. tauto.
exists xB. tauto.
Qed.

Lemma Match3:
  forall (A B C: Prop),
    { A } + { B } + { C } ->
    forall (S: Set) (xA xB xC: S),
      { x: S | (A /\ x = xA) \/ (B /\ x = xB) \/ (C /\ x = xC) }
.

Proof.
intros.
induction H.
induction a.
exists xA. tauto.
exists xB. tauto.
exists xC. tauto.
Qed.

Lemma Match3_diff:
  forall (A B C: Prop),
    { A } + { B } + { C } ->
    (A -> ~B /\ ~C) ->
    (B -> ~A /\ ~C) ->
    (C -> ~A /\ ~B) ->
    forall (S: Set) (xA xB xC: S),
      { x: S |
        (A -> (x = xA)) /\
        (B -> (x = xB)) /\
        (C -> (x = xC))
      }
.

Proof.
intros.
induction H.
induction a.
exists xA. tauto.
exists xB. tauto.
exists xC. tauto.
Qed.

(*****************************************************************)

Lemma match_comp_nat3_aux_1:
  forall (n m: nat),
    (n < m) -> ~(n = m) /\ ~(n > m).

Proof.
intros. omega.
Qed.

Lemma match_comp_nat3_aux_2:
  forall (n m: nat),
    (n = m) -> ~(n < m) /\ ~(n > m).

Proof.
intros. omega.
Qed.

Lemma match_comp_nat3_aux_3:
  forall (n m: nat),
    (n > m) -> ~(n < m) /\ ~(n = m).

Proof.
intros. omega.
Qed.

Definition match_comp_nat3 (n m: nat) (A: Set) (lt_branch eq_branch gt_branch: A): A :=
  match (Match3_diff
    (n < m) (n = m) (n > m) (lt_eq_lt_dec n m)
    (match_comp_nat3_aux_1 n m)
    (match_comp_nat3_aux_2 n m)
    (match_comp_nat3_aux_3 n m)
    A lt_branch eq_branch gt_branch) with
    | (exist x _) => x
  end
.

Lemma match_comp_nat3_lt:
  forall (n m: nat) (A: Set) (lt_branch eq_branch gt_branch: A),
    (n < m) ->
    (match_comp_nat3 n m A lt_branch eq_branch gt_branch) = lt_branch
.

Proof.
intros.
unfold match_comp_nat3.
induction (
  Match3_diff (n < m) (n = m) (n > m) (lt_eq_lt_dec n m)
  (match_comp_nat3_aux_1 n m) (match_comp_nat3_aux_2 n m)
  (match_comp_nat3_aux_3 n m) A lt_branch eq_branch gt_branch
).
tauto.
Qed.

Lemma match_comp_nat3_eq:
  forall (n m: nat) (A: Set) (lt_branch eq_branch gt_branch: A),
    (n = m) ->
    (match_comp_nat3 n m A lt_branch eq_branch gt_branch) = eq_branch
.

Proof.
intros.
unfold match_comp_nat3.
induction (
  Match3_diff (n < m) (n = m) (n > m) (lt_eq_lt_dec n m)
  (match_comp_nat3_aux_1 n m) (match_comp_nat3_aux_2 n m)
  (match_comp_nat3_aux_3 n m) A lt_branch eq_branch gt_branch
).
tauto.
Qed.

Lemma match_comp_nat3_gt:
  forall (n m: nat) (A: Set) (lt_branch eq_branch gt_branch: A),
    (n > m) ->
    (match_comp_nat3 n m A lt_branch eq_branch gt_branch) = gt_branch
.

Proof.
intros.
unfold match_comp_nat3.
induction (
  Match3_diff (n < m) (n = m) (n > m) (lt_eq_lt_dec n m)
  (match_comp_nat3_aux_1 n m) (match_comp_nat3_aux_2 n m)
  (match_comp_nat3_aux_3 n m) A lt_branch eq_branch gt_branch
).
tauto.
Qed.

(*****************************************************************)

Lemma match_comp_nat2_aux_0:
  forall (n m: nat),
    { n < m } + { n >= m }
.

Proof.
intros.
induction (le_lt_dec m n).
right. trivial.
left. trivial.
Qed.

Lemma match_comp_nat2_aux_1:
  forall (n m: nat),
    (n < m) -> ~(n >= m)
.

Proof.
intros. omega.
Qed.

Lemma match_comp_nat2_aux_2:
  forall (n m: nat),
    (n >= m) -> ~(n < m)
.

Proof.
intros. omega.
Qed.

Definition match_comp_nat2 (n m: nat) (A: Set) (lt_branch ge_branch: A): A :=
  match (Match2_diff
    (n < m) (n >= m)
    (match_comp_nat2_aux_0 n m)
    (match_comp_nat2_aux_1 n m)
    (match_comp_nat2_aux_2 n m)
    A lt_branch ge_branch) with
    | (exist x _) => x
  end
.

Lemma match_comp_nat2_lt:
  forall (n m: nat) (A: Set) (lt_branch ge_branch: A),
    (n < m) ->
    (match_comp_nat2 n m A lt_branch ge_branch) = lt_branch
.

Proof.
intros.
unfold match_comp_nat2.
induction (
  Match2_diff (n < m) (n >= m) (match_comp_nat2_aux_0 n m)
  (match_comp_nat2_aux_1 n m) (match_comp_nat2_aux_2 n m) A lt_branch
  ge_branch
).
tauto.
Qed.

Lemma match_comp_nat2_ge:
  forall (n m: nat) (A: Set) (lt_branch ge_branch: A),
    (n >= m) ->
    (match_comp_nat2 n m A lt_branch ge_branch) = ge_branch
.

Proof.
intros.
unfold match_comp_nat2.
induction (
  Match2_diff (n < m) (n >= m) (match_comp_nat2_aux_0 n m)
  (match_comp_nat2_aux_1 n m) (match_comp_nat2_aux_2 n m) A lt_branch
  ge_branch
).
tauto.
Qed.
