(* Author: Vincent Cremet *)
(* Date: July 2006 *) 
(* File: Proofs_AdmissibilityOfTransitivity.v *)
(* Description: Admissibility of transitivity (Appendix 4) *)

(** * Proofs about Admissibility of Transitivity for Subtyping *)

Require Import Calculus.

Module SetProgram(My_Program: Program).

Import My_Program.

(* Import definitions of Proofs_Subclassing.v *)
Require Proofs_Subclassing.
Module My_Proofs_Subclassing := Proofs_Subclassing.SetProgram(My_Program).

Import My_Proofs_Subclassing.
Import My_Proofs_TypeOrdering.
Import My_Proofs_Miscellaneous.
Import My_WellFormedness.
Import My_Semantics.
Import My_Typing.
Import My_TypeOrdering.
Import My_DeBruijn.
Import My_Substitutions.
Import My_Subclassing.

(*****************************************************************)
(*****************************************************************)
(* Unstructuring derivations *)
(*****************************************************************)
(*****************************************************************)

Definition lemmaA23_prop1 (G: Env) (p: Path) (T: type) (H: (PathTyping G p T)): Prop :=
  (Gen_PathTyping G p T)
.

Definition lemmaA23_prop2 (G: Env) (T: type) (U: type) (H: (Subtyping G T U)): Prop :=
  (Gen_Subtyping G T U)
.

Definition lemmaA23_prop3 (G: Env) (C: ClassSym) (fvs: FieldValueList) (H: (FieldValueListTyping G C fvs)): Prop :=
  (Gen_FieldValueListTyping G C fvs)
.

(** Unstructuring derivations **)
Lemma lemmaA23_2:
  forall (G: Env) (T: type) (U: type),
    (Subtyping G T U) ->
    (Gen_Subtyping G T U)
.

Proof.
cut (
  forall (G: Env) (T: type) (U: type) (H: (Subtyping G T U)),
    (lemmaA23_prop2 G T U H)
).
intros. apply H. trivial.
apply (Subtyping_prop
  lemmaA23_prop1
  lemmaA23_prop2
  lemmaA23_prop3
).
(* P-This *)
unfold lemmaA23_prop1.
intros.
eapply Gen_P_This. apply e. apply H.
(* P-Param *)
unfold lemmaA23_prop1.
intros.
eapply Gen_P_Param. apply e. apply i. apply H.
(* P-Var *)
unfold lemmaA23_prop1.
intros.
eapply Gen_P_Var. apply e. apply a. apply H.
(* P-Value *)
unfold lemmaA23_prop1.
intros.
eapply Gen_P_Value. apply l.
eapply H. trivial. apply H0.
(* S-Class *)
unfold lemmaA23_prop2.
intros.
eapply Gen_S_Class.
(* S-Extends *)
unfold lemmaA23_prop2.
intros.
eapply Gen_S_Extends. apply e. apply H.
(* S-Up *)
intros.
unfold lemmaA23_prop2.
eapply Gen_S_Up. apply H. trivial. apply e0. apply H0.
(* S-Down *)
intros.
unfold lemmaA23_prop2.
eapply Gen_S_Down. apply H. trivial. apply e0. apply H0.
(* S-Alias-Left *)
intros.
unfold lemmaA23_prop2.
eapply Gen_S_Alias_Left. apply H. apply e. apply H0.
(* S-Alias-Left *)
intros.
unfold lemmaA23_prop2.
eapply Gen_S_Alias_Right. apply H. apply e. apply H0.
(* S-Virtual *)
intros.
unfold lemmaA23_prop2.
eapply Gen_S_Virtual. apply H. apply e.
(* S-Trans *)
intros.
unfold lemmaA23_prop2.
eapply Gen_S_Trans. apply H. apply H0.
unfold NoConstraint.applyTo.
tauto.
(* Nil (fields-values) *)
intros.
unfold lemmaA23_prop3.
apply Gen_FieldValueListTyping_nil.
(* Cons (fields-values) *)
intros.
unfold lemmaA23_prop3.
eapply Gen_FieldValueListTyping_cons.
apply H. trivial. apply e. trivial. apply H1.
Qed.

(** Unstructuring derivations **)
Lemma lemmaA23_1:
  forall (G: Env) (p: Path) (T: type),
    (PathTyping G p T) ->
    (Gen_PathTyping G p T)
.

Proof.
cut (
  forall (G: Env) (p: Path) (T: type) (H: (PathTyping G p T)),
    (lemmaA23_prop1 G p T H)
).
intros. apply H. trivial.
apply (PathTyping_prop2
  lemmaA23_prop1
  lemmaA23_prop3
).

(* P-This *)
unfold lemmaA23_prop1. intros.
eapply Gen_P_This. apply e. apply lemmaA23_2. trivial.
(* P-Param *)
unfold lemmaA23_prop1. intros.
eapply Gen_P_Param. apply e. apply i. apply lemmaA23_2. trivial.
(* P-Var *)
unfold lemmaA23_prop1. intros.
eapply Gen_P_Var. apply e. apply a. apply lemmaA23_2. trivial.
(* P-Value *)
unfold lemmaA23_prop1. intros.
eapply Gen_P_Value. trivial. apply H. trivial. apply lemmaA23_2. trivial.
(* Nil (fields-values) *)
intros.
unfold lemmaA23_prop3.
apply Gen_FieldValueListTyping_nil.
(* Cons (fields-values) *)
intros.
unfold lemmaA23_prop3.
eapply Gen_FieldValueListTyping_cons.
apply H. trivial. apply e. apply s. apply H0.
Qed.

(*****************************************************************)
(*****************************************************************)
(* Strengthening *)
(*****************************************************************)
(*****************************************************************)

Definition lemmaA24_prop_1 (G: Env) (TA T: type) (H: (Subtyping G TA T)): Prop :=
  forall (C A: ClassSym),
    G = (classEnv C) ->
    TA = (type_class A) ->
    ~(this_in_type T) ->
    (Subtyping rootEnv (type_class A) T)
.

Definition lemmaA24_prop_2 (G: Env) (p: Path) (T: type) (H: (PathTyping G p T)): Prop :=
  forall (C: ClassSym) (v: Value),
    G = (classEnv C) ->
    p = (Path_value v) ->
    ~(this_in_type T) ->
    (PathTyping rootEnv (Path_value v) T)
.

Definition lemmaA24_prop_3 (G: Env) (B: ClassSym) (fvs: FieldValueList) (H: (FieldValueListTyping G B fvs)): Prop :=
  forall (C: ClassSym),
    G = (classEnv C) ->
    (FieldValueListTyping rootEnv B fvs)
.

(* (\** Strengthening - 1 **\) *)
(* Lemma lemmaA24_1: *)
(*   forall (C A: ClassSym) (T: type), *)
(*     (Subtyping (classEnv C) (type_class A) T) -> *)
(*     ~(this_in_type T) -> *)
(*     (Subtyping rootEnv (type_class A) T) *)
(* . *)

(** Strengthening - 2 **)
Lemma lemmaA24_2_aux:
  forall (G: Env) (p: Path) (T: type) (H: PathTyping G p T),
    (lemmaA24_prop_2 G p T H)
.

Proof.
intros.
apply (PathTyping_prop
  lemmaA24_prop_2
  lemmaA24_prop_1
  lemmaA24_prop_3
).
(* Case This *)
unfold lemmaA24_prop_2. intros. discriminate H2.
(* Case Param *)
unfold lemmaA24_prop_2. intros. discriminate H2.
(* Case Var *)
unfold lemmaA24_prop_2. intros. discriminate H2.
(* Case Value *)
intros.
unfold lemmaA24_prop_2. intros.
injection H3; intros. clear H3.
rewrite <- H5.
apply P_Value. trivial.
unfold lemmaA24_prop_3 in H0.
rewrite H2 in H0.
eapply H0 with (C1 := C0). trivial. trivial.
unfold lemmaA24_prop_1 in H1.
rewrite H2 in H1.
eapply H1 with (C1 := C0). trivial. trivial. trivial.

(* Case S-Class *)
unfold lemmaA24_prop_1. intros. 
rewrite H1.
apply (S_Class).
(* Case S-Extends *)
unfold lemmaA24_prop_1. intros.
inversion H2. clear H2.
rewrite <- H5. clear H5.
assert (Subtyping rootEnv (type_class C') T0).
eapply H0. apply H1. trivial. trivial.
eapply S_Extends. apply e. trivial.
(* Case S-Up *)
unfold lemmaA24_prop_1. intros. discriminate H3.
(* Case S-Down *)
unfold lemmaA24_prop_1. intros.
simpl in H4.
assert (Subtyping rootEnv (type_class A) (thisSubst_in_type T0 p0)).
eapply H1. apply H2. trivial.
apply lemmaA24_aux. trivial.
clear H1.
unfold lemmaA24_prop_2 in H0.
rewrite H2 in p1.
induction (path_in_classEnv C0 p0 (type_class C) p1).
rewrite H1 in H4.
red in H4. unfold this_in_path in H4. tauto.
induction H1.
assert (PathTyping rootEnv (Path_value x) (type_class C)).
eapply H0. apply H2. trivial.
unfold not. unfold this_in_type. trivial. clear H0.
eapply S_Down. rewrite H1. apply H6. trivial.
apply e0. trivial.
(* Case S-Alias-Left*)
unfold lemmaA24_prop_1. intros. discriminate H3.
(* Case S-Alias-Right *)
unfold lemmaA24_prop_1. intros.
simpl in H4.
assert (Subtyping rootEnv (type_class A) (thisSubst_in_type T0 p0)).
eapply H1. apply H2. trivial.
apply lemmaA24_aux. trivial.
clear H1.
unfold lemmaA24_prop_2 in H0.
rewrite H2 in p1.
induction (path_in_classEnv C0 p0 (type_class C) p1).
rewrite H1 in H4.
red in H4. unfold this_in_path in H4. tauto.
induction H1.
assert (PathTyping rootEnv (Path_value x) (type_class C)).
eapply H0. apply H2. trivial.
unfold not. unfold this_in_type. trivial. clear H0.
eapply S_Alias_Right. rewrite H1. apply H6.
apply e. trivial.
(* Case S-Virtual *)
unfold lemmaA24_prop_1. intros. discriminate H2.
(* Case S-Trans-S *)
unfold lemmaA24_prop_1.
intros.
simple inversion a. intros. rename H5 into t.
rewrite H3 in t.
induction (lemmaA15_1 S A t).
assert (Subtyping rootEnv (type_class A) S).
eapply H0. apply H2. trivial. rewrite H5. unfold this_in_type. tauto.
assert (Subtyping rootEnv S U).
rewrite H5.
eapply H1. apply H2. trivial. trivial. clear H0 H1.
eapply S_Trans. apply H7. trivial. rewrite <- H3. trivial.

(* Case Nil (field-value) *)
intros.
unfold lemmaA24_prop_3. intros.
apply FieldValueListTyping_nil.
(* Case Cons (field-value) *)
intros.
unfold lemmaA24_prop_3. intros.
unfold lemmaA24_prop_2 in H0.
eapply FieldValueListTyping_cons.
eapply H0. apply H3. trivial.
unfold type_is_closed in t. tauto. trivial. apply e. trivial.
unfold lemmaA24_prop_3 in H2.
eapply H2. apply H3.
Qed.

(** Strengthening - 2 **)
Lemma lemmaA24_2:
  forall (C: ClassSym) (v: Value) (T: type),
    (PathTyping (classEnv C) (Path_value v) T) ->
    ~(this_in_type T) ->
    (PathTyping rootEnv (Path_value v) T)
.

Proof.
intros.
assert (forall (G: Env) (p: Path) (T: type) (H: PathTyping G p T), (lemmaA24_prop_2 G p T H)).
apply lemmaA24_2_aux.
unfold lemmaA24_prop_2 in H1.
eapply H1 with (C := C). apply H. trivial. trivial. trivial.
Qed.

(** Admissibility of transitivity for class type subtyping **)
Lemma lemmaA25:
  forall (G G' G'': Env) (A B C: ClassSym),
    (Subtyping G   (type_class A) (type_class B)) ->
    (Subtyping G'  (type_class B) (type_class C)) ->
    (Subtyping G'' (type_class A) (type_class C))
.

Proof.
intros.
assert (Subclass A B). eapply lemmaA21. apply H.
assert (Subclass B C). eapply lemmaA21. apply H0.
assert (Subclass A C). eapply lemmaA19. apply H1. trivial.
apply lemmaA22. trivial.
Qed.

(*****************************************************************)
(*****************************************************************)
(*****************************************************************)

(** Substitution for self in structured derivations - 1 **)
Definition lemmaA26_prop_1 (G: Env) (p: Path) (T: type) (H: (PathTyping G p T)): Prop :=
  forall (v: Value) (C A: ClassSym),
    (PathTyping rootEnv (Path_value v) (type_class C)) ->
    G = (classEnv C) ->
    T = (type_class A) ->
    (PathTyping rootEnv (thisSubst_in_path p (Path_value v)) (type_class A))
.

(** Substitution for self in structured derivations - 2 **)
Definition lemmaA26_prop_2  (G: Env) (T U: type) (H: (Subtyping G T U)): Prop :=
  forall (v: Value) (C: ClassSym),
    (PathTyping rootEnv (Path_value v) (type_class C)) ->
    G = (classEnv C) ->
    (Subtyping rootEnv (thisSubst_in_type T (Path_value v)) (thisSubst_in_type U (Path_value v)))
.

(** Substitution for self in structured derivations - 1 **)
Lemma lemmaA26_1_aux:
  forall (G: Env) (p: Path) (T: type) (H: (PathTyping G p T)),
    (lemmaA26_prop_1 G p T H).

Proof.
intros G p T H.
unfold lemmaA26_prop_1.
assert (PathTyping G p T). trivial. rename H0 into I. generalize I. clear I.
induction H.

(* Case This *)
intros.
simpl.
rewrite H2 in H. clear H2. unfold classEnv in H. inversion H. clear H.
clear H5 H6.
rewrite H4 in H1. clear H4.
rewrite H3 in H0. clear H3.
inversion H1. clear H1.
assert (Subtyping rootEnv (type_class C1) (type_class A)).
eapply lemmaA25. apply H6. apply H0.
eapply P_Value. trivial. trivial. trivial. trivial.

(* Case Param *)
intros.
rewrite H3 in I.
induction (param_in_classEnv C x S I).

(* Case Var *)
intros.
rewrite H3 in I.
induction (var_in_classEnv C n S I).

(* Case Value *)
intros.
simpl.
rewrite H4 in I.
rewrite H5 in I.
eapply lemmaA24_2. apply I.
simpl. tauto.
Qed.

(** Substitution for self in structured derivations - 1 **)
Lemma lemmaA26_1:
  forall (v: Value) (C: ClassSym),
    (PathTyping rootEnv (Path_value v) (type_class C)) ->
    forall (A: ClassSym) (p: Path),
      (PathTyping (classEnv C) p (type_class A)) ->
      (PathTyping rootEnv (thisSubst_in_path p (Path_value v)) (type_class A))
.

Proof.
intros.
pose (H1 := lemmaA26_1_aux).
unfold lemmaA26_prop_1 in H1.
eapply H1. apply H0. apply H. trivial. trivial.
Qed.

(** Substitution for self in structured derivations - 2 **)
Lemma lemmaA26_2_aux:
  forall (G: Env) (T U: type) (H: (Subtyping G T U)),
    (lemmaA26_prop_2 G T U H).

Proof.
intros G T U H.
unfold lemmaA26_prop_2.
induction H.

(* Case S-Class *)
intros. simpl. apply (S_Class).
(* Case S-Extends *)
intros. simpl. simpl in IHSubtyping.
assert (Subtyping rootEnv (type_class C') (thisSubst_in_type T (Path_value v))).
eapply IHSubtyping. apply H1. trivial.
eapply S_Extends. apply H. trivial.
(* Case S-Up *)
intros. simpl.
assert (PathTyping rootEnv (thisSubst_in_path p (Path_value v)) (type_class C)).
rewrite H4 in H.
eapply (lemmaA26_1). apply H3. trivial.
assert (Subtyping rootEnv
  (thisSubst_in_type (thisSubst_in_type T p) (Path_value v))
  (thisSubst_in_type S (Path_value v))
).
eapply IHSubtyping. apply H3. trivial.
rewrite lemmaA8_2 in H6.
eapply S_Up. apply H5. trivial. apply H1. trivial.
(* Case S-Down *)
intros. simpl.
assert (PathTyping rootEnv (thisSubst_in_path p (Path_value v)) (type_class C)).
rewrite H4 in H.
eapply (lemmaA26_1). apply H3. trivial.
assert (Subtyping rootEnv
  (thisSubst_in_type S (Path_value v))
  (thisSubst_in_type (thisSubst_in_type T p) (Path_value v))
).
eapply IHSubtyping. apply H3. trivial.
rewrite lemmaA8_2 in H6.
eapply S_Down. apply H5. trivial. apply H1. trivial.
(* Case S-Alias-Left *)
intros. simpl.
assert (PathTyping rootEnv (thisSubst_in_path p (Path_value v)) (type_class C)).
rewrite H3 in H.
eapply (lemmaA26_1). apply H2. trivial.
assert (Subtyping rootEnv
  (thisSubst_in_type (thisSubst_in_type T p) (Path_value v))
  (thisSubst_in_type S (Path_value v))
).
eapply IHSubtyping. apply H2. trivial.
rewrite lemmaA8_2 in H5.
eapply S_Alias_Left. apply H4. apply H0. trivial.
(* Case S-Alias-Right *)
intros. simpl.
assert (PathTyping rootEnv (thisSubst_in_path p (Path_value v)) (type_class C)).
rewrite H3 in H.
eapply (lemmaA26_1). apply H2. trivial.
assert (Subtyping rootEnv
  (thisSubst_in_type S (Path_value v))
  (thisSubst_in_type (thisSubst_in_type T p) (Path_value v))
).
eapply IHSubtyping. apply H2. trivial.
rewrite lemmaA8_2 in H5.
eapply S_Alias_Right. apply H4. apply H0. trivial.
(* Case S-Virtual *)
intros. simpl.
assert (PathTyping rootEnv (thisSubst_in_path p (Path_value v)) (type_class C)).
rewrite H2 in H.
eapply (lemmaA26_1). apply H1. trivial.
eapply S_Virtual. apply H3. trivial.
(* Case S-Trans-S *)
intros.
assert (Subtyping rootEnv
  (thisSubst_in_type T (Path_value v))
  (thisSubst_in_type S (Path_value v))
).
eapply IHSubtyping1. apply H2. trivial.
assert (Subtyping rootEnv
  (thisSubst_in_type S (Path_value v))
  (thisSubst_in_type U (Path_value v))
).
eapply IHSubtyping2. apply H2. trivial.
eapply S_Trans. apply H4. apply H5.
destruct H1.
split.
apply lemmaA14. trivial.
apply lemmaA14. trivial.
Qed.

(** Substitution for self in structured derivations - 2 **)
Lemma lemmaA26_2:
  forall (v: Value) (C: ClassSym) (T U: type),
    (PathTyping rootEnv (Path_value v) (type_class C)) ->
    (Subtyping (classEnv C) T U) ->
    (Subtyping rootEnv (thisSubst_in_type T (Path_value v)) (thisSubst_in_type U (Path_value v)))
.

Proof.
intros.
pose (H1 := lemmaA26_2_aux).
unfold lemmaA26_prop_2 in H1.
eapply H1. apply H0. apply H. trivial.
Qed.

(*************************************************************)
(*************************************************************)
(*************************************************************)
(*************************************************************)
(* Admissibility of transitivity for structured subtyping *)

Import Type_Multiset.
Import Type_MultisetOrder.
Import MSet.

Definition lemmaA27_prop (tm: Type_Multiset.Multiset): Prop :=
  forall (T S U: type),
    tm = ({{T}} + {{S}} + {{U}}) ->
    (Subtyping rootEnv T S) ->
    (Subtyping rootEnv S U) ->
    (Subtyping rootEnv T U)
.

(** Admissibility of transitivity for structured subtyping **)
Lemma lemmaA27:
  forall (T S U: type),
    (Subtyping rootEnv T S) ->
    (Subtyping rootEnv S U) ->
    (Subtyping rootEnv T U)
.

Proof.
cut (forall (tm: Type_Multiset.Multiset), (lemmaA27_prop  tm)).
intros. unfold lemmaA27_prop in H.
pose (tm := {{T}} + {{S}} + {{U}}).
eapply H with (S := S). assert (tm =  {{T}} + {{S}} + {{U}}).
trivial. apply H2. trivial. trivial.

apply well_founded_ind with (R := typeRelMul).
apply lemmaA18.
unfold lemmaA27_prop.
intros.
rewrite H0 in H. clear H0 x.

pose (G := rootEnv). assert (G = rootEnv). trivial.
replace (rootEnv) with G in H1.
destruct H1.

(* Case (S-Class,?) *)
clear H0 G. trivial.

(* Case (S-Extends,?) *)
rewrite H0 in H3. clear H0 G.
assert (typeRel (type_class C') (type_class C)).
apply lemmaA13_3. trivial.
assert (typeRelMul
  ({{type_class C'}} + {{T}} + {{U}})
  ({{type_class C}} + {{T}} + {{U}})
).
apply typeRelMul1. trivial.
assert (Subtyping rootEnv (type_class C') U).
eapply H with (S := T).
apply H4. trivial. trivial. trivial.
eapply S_Extends. apply H1. trivial.

(* Case (S-Up,?) *)
rewrite H0 in H1. rewrite H0 in H5. clear H0 G.
assert (typeRel (thisSubst_in_type T p) (type_virtual p L)).
apply lemmaA13_1. trivial.
assert (typeRelMul
  ({{(thisSubst_in_type T p)}} + {{S}} + {{U}})
  ({{type_virtual p L}} + {{S}} + {{U}})
).
apply typeRelMul1. trivial.
assert (Subtyping rootEnv ((thisSubst_in_type T p)) U).
eapply H with (S := S).
apply H6. trivial. trivial. trivial.
eapply S_Up. apply H1. trivial. apply H4. trivial.

Focus 2.
(* Case (S-Alias-Left,?) *)
rewrite H0 in H1. rewrite H0 in H4. clear H0 G.
assert (typeRel (thisSubst_in_type T p) (type_virtual p L)).
eapply lemmaA13_2. apply H3.
assert (typeRelMul
  ({{(thisSubst_in_type T p)}} + {{S}} + {{U}})
  ({{type_virtual p L}} + {{S}} + {{U}})
).
apply typeRelMul1. trivial.
assert (Subtyping rootEnv ((thisSubst_in_type T p)) U).
eapply H with (S := S).
apply H5. trivial. trivial. trivial.
eapply S_Alias_Left. apply H1. apply H3. trivial.

Focus 3.
(* Case (S-Virtual,?) *)
rewrite H0 in H1.  clear H0 G.
trivial.

Focus 3.
(* Case (S-Trans-S,?) *)
rewrite H0 in H1_. rewrite H0 in H1_0. clear H0 G.
rename S into S'. rename U0 into S.
inversion H1.
assert (typeRelMul
  ({{S'}} + {{S}} + {{U}})
  ({{T}} + {{S}} + {{U}})
).
apply typeRelMul1. trivial.
assert (Subtyping rootEnv S' U).
eapply H with (S := S). apply H4. trivial. trivial. trivial.
assert (typeRelMul
  ({{T}} + {{S'}} + {{U}})
  ({{T}} + {{S}} + {{U}})
).
apply typeRelMul3. trivial.
eapply H with (S := S'). apply H6. trivial. trivial. trivial.

(*************************************************************************)
(* Case (S-Down,...) *)
(*************************************************************************)

rename T into T'. rename S into T.
pose (S := (type_virtual p L)). assert (S = (type_virtual p L)). trivial.
replace (type_virtual p L) with S in H2.
replace (rootEnv) with G in H2.
destruct H2.

(* Case (S-Down,S-Class) *)
discriminate H6.

(* Case (S-Down,S-Extends) *)
discriminate H6.

(* Case (S-Down,S-Up) *)
(* start: clean up *)
rewrite H0 in H1. rewrite H0 in H5. rewrite H0 in H2. rewrite H0 in H9. clear H0 G.
inversion H6. clear H6. rewrite H10 in H2. rewrite H10 in H9. clear H10 p0.
rewrite H11 in H7. rewrite H11 in H8. clear H11 L0.
rewrite H3 in H7. rewrite <- H7 in H2. clear H7 C0.
rename T0 into U'. clear H2. rename S into U.
(* end: clean up *)

(* p = v,  v = new A(fvs) *)
induction (lemmaA7 p (type_class C) H1). rename x into v.
rewrite H0 in H1.
inversion H1. rename C0 into A. clear H11 G. clear H13 S.

(* A subclass of C *)
assert (Subclass A C). eapply lemmaA21. apply H12.

(* (type L = S') in C',  A subclass of C' *)
inversion H10. clear H16 C0. clear H17 fs. clear H13 H14.
unfold IsComplete_types in H15.
induction (H15 C L H11 H3). clear H15. induction H13. induction H13.
rename x into C'. rename x0 into S'.

(* this: C' |-  S' <: U',  this: C' |- T' S',   C' subclass of C *)
assert (Subtyping (classEnv C') S' U').
eapply typeVal_subtype_of_upperBound with (L := L). trivial. trivial.
assert (Subtyping (classEnv C') T' S').
apply typeVal_supertype_of_lowerBound with (L := L). trivial. trivial.

(* A subclass of C' *)
assert (Subtyping rootEnv (type_class A) (type_class C')). apply lemmaA22. trivial.

(* |- v: C' *)
assert (PathTyping rootEnv (Path_value v) (type_class C')).
rewrite <- H2.
eapply P_Value. trivial. trivial. trivial. trivial.

(* |- S'[v] <: U'[v] *)
assert (Subtyping rootEnv (thisSubst_in_type S' (Path_value v)) (thisSubst_in_type U' (Path_value v))).
eapply lemmaA26_2. apply H18. trivial.
rewrite <- H0 in H19.

(* |- T'[v] <: S'[v] *)
assert (Subtyping rootEnv (thisSubst_in_type T' (Path_value v)) (thisSubst_in_type S' (Path_value v))).
eapply lemmaA26_2. apply H18. trivial.
rewrite <- H0 in H20.

(* S'[p] << p.L *)
assert (typeRel (thisSubst_in_type S' p) (type_virtual p L)).
eapply lemmaA13_2. apply H14.

(* U'[p] << p.L *)
assert (typeRel (thisSubst_in_type U' p) (type_virtual p L)).
eapply lemmaA13_1. trivial.

(* T'[p] << p.L *)
assert (typeRel (thisSubst_in_type T' p) (type_virtual p L)).
eapply lemmaA13_1'. trivial.

(* { T, T'[p], S'[p] } << { T, p.L, U } *)
assert (typeRelMul
  ({{T}} + {{(thisSubst_in_type T' p)}} + {{(thisSubst_in_type S' p)}})
  ({{T}} + {{type_virtual p L}} + {{U}})
).
apply typeRelMul2. trivial. trivial.

(* |- T <: S'[p] *)
assert ( Subtyping rootEnv T (thisSubst_in_type S' p)).
eapply H with (S := (thisSubst_in_type T' p)). apply H24. trivial. trivial. trivial.

(* { T, S'[p], U'[p] } << { T, p.L, U } *)
assert (typeRelMul
  ({{T}} + {{(thisSubst_in_type S' p)}} + {{(thisSubst_in_type U' p)}})
  ({{T}} + {{type_virtual p L}} + {{U}})
).
apply typeRelMul2. trivial. trivial.

(* |- T <: U'[p] *)
assert ( Subtyping rootEnv T (thisSubst_in_type U' p)).
eapply H with (S := (thisSubst_in_type S' p)). apply H26. trivial. trivial. trivial.

(* { T, U'[p], U } << { T, p.L, U } *)
assert (typeRelMul
  ({{T}} + {{(thisSubst_in_type U' p)}} + {{U}})
  ({{T}} + {{type_virtual p L}} + {{U}})
).
apply typeRelMul3. trivial.

(* |- T <: U *)
eapply H with (S := (thisSubst_in_type U' p)). apply H28. trivial. trivial. trivial.

(* Case (S-Down,S-Down) *)
(* start: clean up *)
rewrite H0 in H1. rewrite H0 in H5. rewrite H0 in H9.
rewrite H0 in H2. clear H0 G.
(* end: clean up *)
assert (typeRel (thisSubst_in_type T0 p0) (type_virtual p0 L0)).
apply lemmaA13_1'. trivial.
assert (typeRelMul
  ({{T}} + {{S}} + {{(thisSubst_in_type T0 p0)}})
  ({{T}} + {{S}} + {{(type_virtual p0 L0)}})
).
apply typeRelMul4. trivial.
assert (Subtyping rootEnv T ((thisSubst_in_type T0 p0))).
eapply H with (S := S). rewrite <- H6.
apply H10. trivial.
rewrite H6. eapply S_Down. apply H1. trivial. apply H4. trivial. trivial.
eapply S_Down. apply H2. trivial. apply H8. trivial.

(* Case (S-Down,S-Alias-Left) *)
(* start: clean up *)
rewrite H0 in H1. rewrite H0 in H2. rewrite H0 in H8.
rewrite H0 in H5. clear H0 G.
inversion H6. clear H6.
rewrite H9 in H2. rewrite H9 in H8. clear H9 p0.
rewrite H10 in H7. clear H10 L0.
rename T0 into U'. rename C into C'. rename C0 into C. rename S into U.
(* end: clean up *)

(* this: C |-  T' <: U',  C subclass of C' *)
assert (Subtyping (classEnv C) T' U').
eapply typeVal_supertype_of_lowerBound with (L := L). trivial. trivial.

(* p = v *)
induction (lemmaA7 p (type_class C) H2). rename x into v.
rewrite H6 in H2.

(* |- T'[v] <: U'[v] *)
assert (Subtyping rootEnv (thisSubst_in_type T' (Path_value v)) (thisSubst_in_type U' (Path_value v))).
eapply lemmaA26_2. apply H2. trivial.
rewrite <- H6 in H9.

(* U'[p] << p.L *)
assert (typeRel (thisSubst_in_type U' p) (type_virtual p L)).
eapply lemmaA13_2. apply H7.

(* T'[p] << p.L *)
assert (typeRel (thisSubst_in_type T' p) (type_virtual p L)).
eapply lemmaA13_1'. trivial.

(* { T, T'[p], U'[p] } << { T, p.L, U } *)
assert (typeRelMul
  ({{T}} + {{(thisSubst_in_type T' p)}} + {{(thisSubst_in_type U' p)}})
  ({{T}} + {{type_virtual p L}} + {{U}})
).
apply typeRelMul2. trivial. trivial.

(* |- T <: U'[p] *)
assert (Subtyping rootEnv T (thisSubst_in_type U' p)).
eapply H with (S := (thisSubst_in_type T' p)). apply H12. trivial. trivial. trivial.

(* { T, U'[p], U } << { T, p.L, U } *)
assert (typeRelMul
  ({{T}} + {{(thisSubst_in_type U' p)}} + {{U}})
  ({{T}} + {{type_virtual p L}} + {{U}})
).
apply typeRelMul3. trivial.

(* |- T <: U *)
eapply H with (S := (thisSubst_in_type U' p)). apply H14. trivial. trivial. trivial.

(* Case (S-Down,S-Alias-Right) *)
(* start: clean up *)
rewrite H0 in H1. rewrite H0 in H5. rewrite H0 in H2.
rewrite H0 in H8. clear H0 G.
(* end: clean up *)
assert (typeRel (thisSubst_in_type T0 p0) (type_virtual p0 L0)).
eapply lemmaA13_2. apply H7.
assert (typeRelMul
  ({{T}} + {{S}} + {{(thisSubst_in_type T0 p0)}})
  ({{T}} + {{S}} + {{(type_virtual p0 L0)}})
).
apply typeRelMul4. trivial.
assert (Subtyping rootEnv T ((thisSubst_in_type T0 p0))).
eapply H with (S := S). rewrite <- H6.
apply H9. trivial.
rewrite H6. eapply S_Down. apply H1. trivial. apply H4. trivial. trivial.
eapply S_Alias_Right. apply H2. apply H7. trivial.

(* Case (S-Down,S-Virtual) *)
rewrite H0 in H1. rewrite H0 in H5. rewrite H0 in H2. clear H0 G.
inversion H6.
eapply S_Down. apply H1. trivial. apply H4. apply H5.

(* Case (S-Down,S-Trans) *)
rewrite H0 in H1. rewrite H0 in H5. rewrite H0 in H2_. rewrite H0 in H2_0. clear H0 G.
rename S into S'. rename T0 into S.
assert (Subtyping rootEnv T S). rewrite H6.
eapply S_Down. apply H1. trivial. apply H4. trivial.
inversion H2.
assert (typeRelMul
  ({{T}} + {{S}} + {{S'}})
  ({{T}} + {{S}} + {{U}})
).
apply typeRelMul4. trivial.
assert (Subtyping rootEnv T S').
eapply H with (S := S). rewrite <- H6. apply H9. trivial. trivial. trivial.
assert (typeRelMul
  ({{T}} + {{S'}} + {{U}})
  ({{T}} + {{S}} + {{U}})
).
apply typeRelMul3. trivial.
eapply H with (S := S'). rewrite <- H6. apply H11. trivial. trivial. trivial.

(*************************************************************************)
(* Case (S-Alias-Right,...) *)
(*************************************************************************)

rename T into T'. rename S into T.
pose (S := (type_virtual p L)). assert (S = (type_virtual p L)). trivial.
replace (type_virtual p L) with S in H2.
replace (rootEnv) with G in H2.
destruct H2.

(* Case (S-Alias-Right,S-Class) *)
discriminate H5.

(* Case (S-Alias-Right,S-Extends) *)
discriminate H5.

(* Case (S-AliasRight,S-Up) *)
(* start: clean up *)
rewrite H0 in H1. rewrite H0 in H4. rewrite H0 in H2.
rewrite H0 in H8. clear H0 G.
inversion H5. clear H5.
rewrite H9 in H2. rewrite H9 in H8. clear H9 p0.
rewrite H10 in H7. rewrite H10 in H6. clear H10 L0.
rename T0 into U'. rename C0 into C'. rename S into U.
(* end: clean up *)

(* this: C |-  T' <: U' *)
assert (Subtyping (classEnv C) T' U').
eapply typeVal_subtype_of_upperBound with (L := L). trivial. trivial.

(* p = v *)
induction (lemmaA7 p (type_class C) H1). rename x into v.
rewrite H5 in H1.

(* |- T'[v] <: U'[v] *)
assert (Subtyping rootEnv (thisSubst_in_type T' (Path_value v)) (thisSubst_in_type U' (Path_value v))).
eapply lemmaA26_2. apply H1. trivial.
rewrite <- H5 in H9.

(* T'[p] << p.L *)
assert (typeRel (thisSubst_in_type T' p) (type_virtual p L)).
eapply lemmaA13_2. apply H3.

(* U'[p] << p.L *)
assert (typeRel (thisSubst_in_type U' p) (type_virtual p L)).
eapply lemmaA13_1. trivial.

(* { T, T'[p], U'[p] } << { T, p.L, U } *)
assert (typeRelMul
  ({{T}} + {{(thisSubst_in_type T' p)}} + {{(thisSubst_in_type U' p)}})
  ({{T}} + {{type_virtual p L}} + {{U}})
).
apply typeRelMul2. trivial. trivial.

(* |- T <: U'[p] *)
assert (Subtyping rootEnv T (thisSubst_in_type U' p)).
eapply H with (S := (thisSubst_in_type T' p)). apply H12. trivial. trivial. trivial.

(* { T, U'[p], U } << { T, p.L, U } *)
assert (typeRelMul
  ({{T}} + {{(thisSubst_in_type U' p)}} + {{U}})
  ({{T}} + {{type_virtual p L}} + {{U}})
).
apply typeRelMul3. trivial.

(* |- T <: U *)
eapply H with (S := (thisSubst_in_type U' p)). apply H14. trivial. trivial. trivial.

(* Case (S-Alias-Right,S-Down) *)
(* start: clean up *)
rewrite H0 in H1. rewrite H0 in H4. rewrite H0 in H2.
rewrite H0 in H8. clear H0 G.
(* end: clean up *)
assert (typeRel (thisSubst_in_type T0 p0) (type_virtual p0 L0)).
apply lemmaA13_1'. trivial.
assert (typeRelMul
  ({{T}} + {{S}} + {{(thisSubst_in_type T0 p0)}})
  ({{T}} + {{S}} + {{(type_virtual p0 L0)}})
).
apply typeRelMul4. trivial.
assert (Subtyping rootEnv T ((thisSubst_in_type T0 p0))).
eapply H with (S := S). rewrite <- H5.
apply H9. trivial.
rewrite H5. eapply S_Alias_Right. apply H1. apply H3. trivial. trivial.
eapply S_Down. apply H2. trivial. apply H7. trivial.

(* Case (S-Alias-Right,S-Alias-Left) *)
(* start: clean up *)
rewrite H0 in H1. rewrite H0 in H4. rewrite H0 in H2.
rewrite H0 in H7. clear H0 G.
inversion H5. clear H5.
rewrite H8 in H2. rewrite H8 in H7. clear H8 p0.
rewrite H9 in H6. clear H9 L0.
rename T0 into U'. rename C0 into C'. rename S into U.
(* end: clean up *)

(* p = v,   v = new A(fvs) *)
induction (lemmaA7 p (type_class C) H1). rename x into v.
rewrite H0 in H1.
inversion H1. rename C0 into A. clear H11 H13 S.

(* A subclass of C *)
assert (Subclass A C).
eapply lemmaA21. apply H12.

(* p = v,   v = new A(fvs) *)
rewrite H0 in H2.
inversion H2. clear H17 H19 G0 S.
rewrite <- H5 in H13. injection H13; intros. clear H13.
rewrite H19 in H18.

(* A subclass of C' *)
assert (Subclass A C').
eapply lemmaA21. apply H18.

(* C subclass of C' or C' subclass of C *)
assert ((Subclass C C') \/ (Subclass C' C)).
eapply lemmaA20. apply H11. trivial.

(* (T' = U') and (C = C') *)
assert ((T' = U') /\ (C = C')).
induction H20.
eapply no_type_overriding_proof. apply H3. trivial. trivial.
assert ((U' = T') /\ (C' = C)).
eapply no_type_overriding_proof. apply H6. trivial. trivial.
induction H21. split. auto. auto.
induction H21.
rewrite <- H21 in H7.

(* T'[p] << p.L *)
assert (typeRel (thisSubst_in_type T' p) (type_virtual p L)).
eapply lemmaA13_2. apply H3.

(* { T, T'[p], U } << { T, p.L, U } *)
assert (typeRelMul
  ({{T}} + {{(thisSubst_in_type T' p)}} + {{U}})
  ({{T}} + {{type_virtual p L}} + {{U}})
).
apply typeRelMul3. trivial.

(* |- T <: U *)
eapply H with (S := (thisSubst_in_type T' p)). apply H24. trivial. trivial. trivial.

(* Case (S-Alias-Right,S-Alias-Right) *)
(* start: clean up *)
rewrite H0 in H1. rewrite H0 in H4. rewrite H0 in H2.
rewrite H0 in H7. clear H0 G.
(* end: clean up *)
assert (typeRel (thisSubst_in_type T0 p0) (type_virtual p0 L0)).
eapply lemmaA13_2. apply H6.
assert (typeRelMul
  ({{T}} + {{S}} + {{(thisSubst_in_type T0 p0)}})
  ({{T}} + {{S}} + {{(type_virtual p0 L0)}})
).
apply typeRelMul4. trivial.
assert (Subtyping rootEnv T ((thisSubst_in_type T0 p0))).
eapply H with (S := S). rewrite <- H5.
apply H8. trivial.
rewrite H5. eapply S_Alias_Right. apply H1. apply H3. trivial. trivial.
eapply S_Alias_Right. apply H2. apply H6. trivial.

(* Case (S-Alias-Right,S-Virtual) *)
rewrite H0 in H1. rewrite H0 in H4. rewrite H0 in H2. clear H0 G.
inversion H5.
eapply S_Alias_Right. apply H1. apply H3. apply H4.

(* Case (S-alias-Right,S-Trans-S) *)
rewrite H0 in H1. rewrite H0 in H4. rewrite H0 in H2_. rewrite H0 in H2_0. clear H0 G.
rename S into S'. rename T0 into S.
assert (Subtyping rootEnv T S). rewrite H5.
eapply S_Alias_Right. apply H1. apply H3. trivial.
inversion H2.
assert (typeRelMul
  ({{T}} + {{S}} + {{S'}})
  ({{T}} + {{S}} + {{U}})
).
apply typeRelMul4. trivial.
assert (Subtyping rootEnv T S').
eapply H with (S := S). rewrite <- H5. apply H8. trivial. trivial. trivial.
assert (typeRelMul
  ({{T}} + {{S'}} + {{U}})
  ({{T}} + {{S}} + {{U}})
).
apply typeRelMul3. trivial.
eapply H with (S := S'). rewrite <- H5. apply H10. trivial. trivial. trivial.
Qed.

(*****************************************************************)
(*****************************************************************)
(* Structuring derivations *)
(*****************************************************************)
(*****************************************************************)

Definition lemmaA28_prop1 (G: Env) (p: Path) (T: type) (H: (Gen_PathTyping G p T)): Prop :=
  forall (v: Value),
    (G = rootEnv) ->
    (p = (Path_value v)) ->
    (PathTyping rootEnv p T)
.

Definition lemmaA28_prop2 (G: Env) (T: type) (U: type) (H: (Gen_Subtyping G T U)): Prop :=
  (G = rootEnv) ->
  (Subtyping rootEnv T U)
.

Definition lemmaA28_prop3 (G: Env) (C: ClassSym) (fvs: FieldValueList) (H: (Gen_FieldValueListTyping G C fvs)): Prop :=
  (G = rootEnv) ->
  (FieldValueListTyping rootEnv C fvs)
.

(** Structuring derivations -- subtyping **)
Lemma lemmaA28_2:
  forall (T: type) (U: type),
    (Gen_Subtyping rootEnv T U) ->
    (Subtyping rootEnv T U)
.

Proof.
cut (
  forall (G: Env) (T: type) (U: type) (H: (Gen_Subtyping G T U)),
    (lemmaA28_prop2 G T U H)
).
intros. unfold lemmaA28_prop2 in H. eapply H. apply H0. trivial.
apply (Gen_Subtyping_prop
  lemmaA28_prop1
  lemmaA28_prop2
  lemmaA28_prop3
).
(* P-This *)
intros.
unfold lemmaA28_prop1.
intros.
discriminate H1.
(* P-Param *)
intros.
unfold lemmaA28_prop1.
intros.
discriminate H1.
(* P-Var *)
intros.
unfold lemmaA28_prop1.
intros.
discriminate H1.
(* P-Value *)
intros.
unfold lemmaA28_prop1.
intros.
eapply P_Value. trivial. apply H. trivial. trivial.
apply H0. trivial.
(* S-Class *)
intros.
unfold lemmaA28_prop2.
intros.
eapply S_Class.
(* S-Extends *)
intros.
unfold lemmaA28_prop2.
intros.
eapply S_Extends. apply e. apply H. trivial.
(* S-Up *)
unfold lemmaA28_prop1.
unfold lemmaA28_prop2.
intros.
rewrite H1 in g.
destruct (lemmaA7' p (type_class C) g).
eapply S_Up.
eapply H. trivial. apply H2. trivial. apply e0.
eapply H0. trivial.
(* S-Down *)
unfold lemmaA28_prop1.
unfold lemmaA28_prop2.
intros.
rewrite H1 in g.
destruct (lemmaA7' p (type_class C) g).
eapply S_Down.
eapply H. trivial. apply H2. trivial. apply e0.
eapply H0. trivial.
(* S-Alias-Left *)
unfold lemmaA28_prop1.
unfold lemmaA28_prop2.
intros.
rewrite H1 in g.
destruct (lemmaA7' p (type_class C) g).
eapply S_Alias_Left.
eapply H. trivial. apply H2. apply e.
eapply H0. trivial.
(* S-Alias-Right *)
unfold lemmaA28_prop1.
unfold lemmaA28_prop2.
intros.
rewrite H1 in g.
destruct (lemmaA7' p (type_class C) g).
eapply S_Alias_Right.
eapply H. trivial. apply H2. apply e.
eapply H0. trivial.
(* S-Virtual *)
unfold lemmaA28_prop1.
unfold lemmaA28_prop2.
intros.
rewrite H0 in g.
destruct (lemmaA7' p (type_class C) g).
eapply S_Virtual.
eapply H. trivial. apply H1. trivial.
(* S-Trans-S *)
unfold lemmaA28_prop2.
intros.
eapply lemmaA27.
apply H. trivial. apply H0. trivial.
(* Nil (field-value) *)
intros.
unfold lemmaA28_prop3.
intros.
apply FieldValueListTyping_nil.
(* Cons (field-value) *)
intros.
unfold lemmaA28_prop3.
intros.
eapply FieldValueListTyping_cons.
unfold lemmaA28_prop1 in H.
eapply H with (v0 := v). trivial. trivial. trivial.
apply e.
exact s.
apply H0. trivial.
Qed.

(** Structuring derivations -- path typing **)
Lemma lemmaA28_1:
  forall (v: Value) (T: type),
    (Gen_PathTyping rootEnv (Path_value v) T) ->
    (PathTyping rootEnv (Path_value v) T)
.

Proof.
cut (
  forall (G: Env) (p: Path) (T: type) (H: (Gen_PathTyping G p T)),
    (lemmaA28_prop1 G p T H)
).
intros.
unfold lemmaA28_prop1 in H.
eapply H with (v := v). apply H0. trivial. trivial.
apply (Gen_PathTyping_prop2
  lemmaA28_prop1
  lemmaA28_prop3
).

(* P-This *)
intros.
unfold lemmaA28_prop1. intros.
discriminate H0.
(* P-Param *)
intros.
unfold lemmaA28_prop1. intros.
discriminate H0.
(* P-Var *)
intros.
unfold lemmaA28_prop1. intros.
discriminate H0.
(* P-Value *)
intros.
unfold lemmaA28_prop1. intros.
eapply P_Value. trivial. apply H. trivial.
trivial. apply lemmaA28_2. rewrite <- H0. trivial.
(* Nil (field-value) *)
intros. unfold lemmaA28_prop3. intros.
apply FieldValueListTyping_nil.
(* Cons (field-value) *)
intros. unfold lemmaA28_prop3. intros.
eapply FieldValueListTyping_cons.
unfold lemmaA28_prop1 in H.
eapply H with (v0 := v). trivial. trivial.
trivial. apply e. trivial.
apply H0. trivial.
Qed.

End SetProgram.

