(* Author: Vincent Cremet *)
(* Date: July 2006 *) 
(* File: Proofs_Subclassing.v *)
(* Description: Properties of the subclassing relation (Appendix 3) *)

(** * Proofs about Subclassing *)

Require Import Calculus.
Require Import Arith.

Module SetProgram(My_Program: Program).

Import My_Program.

(* Import definitions of Proofs_TypeOrdering.v *)
Require Proofs_TypeOrdering.
Module My_Proofs_TypeOrdering := Proofs_TypeOrdering.SetProgram(My_Program).

Import My_Proofs_TypeOrdering.
Import My_Proofs_Miscellaneous.
Import My_WellFormedness.
Import My_Semantics.
Import My_Typing.
Import My_TypeOrdering.
Import My_DeBruijn.
Import My_Substitutions.
Import My_Subclassing.

Hint Constructors Subclass.

(** Transitivity of subclassing **)
Lemma lemmaA19: forall (A B C: ClassSym),
  (Subclass A B) ->
  (Subclass B C) ->
  (Subclass A C).

Proof.
intros.
induction H.
eauto.
eauto.
Qed.

(** Transitivity of subclassing **)
Lemma lemmaA20: forall (C C1 C2: ClassSym),
  (Subclass C C1) ->
  (Subclass C C2) ->
  (Subclass C1 C2) \/ (Subclass C2 C1).

Proof.
intros C C1 C2 H.
generalize C2.
induction H.
tauto.
intros C3 H'.
inversion H'.
right.
rewrite <- H2.
eauto.
apply IHSubclass.
assert (Some B = Some B0).
rewrite <- H.
rewrite <- H1.
trivial.
inversion H5.
trivial.
Qed.

(** Subclassing implies subtyping ***)
Lemma lemmaA22:
  forall (C C': ClassSym) (G: Env),
    (Subclass C C') ->
    (Subtyping G (type_class C) (type_class C'))
.

Proof.
intros C C' G H.
induction H.
apply (S_Class).
apply (S_Extends G A B (type_class C)).
trivial. trivial.
Qed.

(** Subtyping implies subclassing ***)
Lemma lemmaA21:
  forall (C C': ClassSym) (G: Env),
    (Subtyping G (type_class C) (type_class C')) ->
    (Subclass C C')
.

Proof.
  Lemma lemmaA21_aux:
    forall (G: Env) (T U: type),
      (Subtyping G T U) ->
      forall (C C': ClassSym),
        T = (type_class C) ->
        U = (type_class C') ->
        (Subclass C C')
  .

  Proof.
    intros G T U H.
    induction H.

    (* S_Class *)
    intros.
    inversion H.
    inversion H0.
    rewrite <- H2.
    rewrite <- H3. 
    apply SC_Class.

    (* S_Extends *)
    intros.
    inversion H1.
    rewrite <- H4.
    assert (Subclass C' C'0).
    apply (IHSubtyping C' C'0).
    trivial. trivial.    
    eapply (SC_Extends).
    apply H. trivial.

    (* Various simple cases *)
    intros. discriminate H3.
    intros. discriminate H4.
    intros. discriminate H2.
    intros. discriminate H3.
    intros. discriminate H1.
    
    (* S_Trans *)
    intros.
    unfold SymbolConstraint.applyTo in H1.
    destruct H1.
    rewrite H2 in H1. 
    rewrite H3 in H4.
    induction (lemmaA15_1 S C H1).
    rename x into C''.
    assert (Subclass C C'').
    apply (IHSubtyping1). trivial. trivial.
    assert (Subclass C'' C').
    apply (IHSubtyping2). trivial. trivial.
    eapply lemmaA19. apply H6. trivial.
    Qed.
    
(* Show. *)
intros.
eapply lemmaA21_aux. apply H. trivial. trivial.
Qed.

End SetProgram.
