(* Author: Vincent Cremet *)
(* Date: August 2006 *) 
(* File: Proofs_SubjectReduction.v *)
(* Description: Proof of subject-reduction (Appendix 8) *)

(** * Proof of the Subject Reduction Property *)

Require Import Calculus.
Require Import Arith.
Require Import MatchCompNat.
Require Import Util.
Require Import List.
Require Import ListSet.

Module SetProgram(My_Program: Program).

Import My_Program.

(* Import definitions of Proofs_SubstitutionLemmas.v *)
Require Proofs_SubstitutionLemmas3.
Module My_Proofs_SubstitutionLemmas3 := Proofs_SubstitutionLemmas3.SetProgram(My_Program).

Import My_Proofs_SubstitutionLemmas3.
Import My_Proofs_SubstitutionLemmas2.
Import My_Proofs_SubstitutionLemmas.
Import My_Proofs_DeBruijnSIndices.
Import My_Proofs_Progress.
Import My_Proofs_AdmissibilityOfTransitivity.
Import My_Proofs_Subclassing.
Import My_Proofs_TypeOrdering.
Import My_Proofs_Miscellaneous.
Import My_WellFormedness.
Import My_Semantics.
Import My_Typing.
Import My_TypeOrdering.
Import My_DeBruijn.
Import My_Substitutions.
Import My_Subclassing.

(*****************************************************************)

Lemma lemmaA40_aux_1:
  forall (G: Env) (C: ClassSym) (fvs: FieldValueList)
    (f: FieldSym) (v: Value) (T: type),
    (Gen_FieldValueListTyping G C fvs) ->
    (FieldValueList_contains fvs f v) ->
    (fieldType f) = T ->
    exists U: type,
      (Gen_PathTyping G (Path_value v) U) /\
      (type_is_closed U) /\      
      (Subtyping (classEnv C) U T)
.

Proof.
intros.
generalize v H0. clear v H0.
induction fvs.
(* Nil *)
intros.
inversion H0.
intros.
inversion_clear H.
inversion H0.
(* Cons - head *)
clear f1 v1 fvs0 H7 H8 H9.
exists U.
rewrite H10 in H4. rewrite H1 in H4. rewrite <- H4 in H5.
rewrite H11 in H2.
tauto.
(* Cons - tail *)
apply IHfvs. trivial. trivial.
Qed.

Lemma lemmaA40_aux_2:
  forall (v: Value),
    (thisSubst_in_paramTypeList nil (Path_value v)) = nil
.

Proof.
intros. simpl. trivial.
Qed.

Lemma lemmaA40_aux_3:
  forall (v: Value),
    (thisSubst_in_typeList nil (Path_value v)) = nil
.

Proof.
intros. simpl. trivial.
Qed.

(******************************************************************)

Lemma lemmaA40_aux_5:
  forall (xts: ParamTermList) (xts': ParamTermList),
    (Red_paramTermList xts xts') ->
    (params_of_paramTermList xts) = (params_of_paramTermList xts')
.

Proof.
induction xts.
intros. inversion H.
intros. simpl.
destruct xts'. inversion H.
inversion_clear H.
simpl. trivial.
simpl.
rewrite (IHxts xts' H1). trivial.
Qed.

Lemma lemmaA40_aux_16:
  forall (fts: FieldTermList) (fts': FieldTermList),
    (Red_fieldTermList fts fts') ->
    (fields_of_fieldTermList fts) = (fields_of_fieldTermList fts')
.

Proof.
induction fts.
intros. inversion H.
intros. simpl.
destruct fts'. inversion H.
inversion_clear H.
simpl. trivial.
simpl.
rewrite (IHfts fts' H1). trivial.
Qed.

Lemma lemmaA40_aux_6:
  forall (x: ParamSym) (T Tp: type) (m: MethodSym) (p: Path),
    (In (x, Tp) (thisSubst_in_paramTypeList (method_parameters m) p)) ->
    (paramType x) = T ->
    (paramOwner x) = m /\ Tp = (thisSubst_in_type T p)
.

Proof.
intros.
destruct (map_elim _ _ _ _ _ H).
destruct x0. destruct H1.
injection H2; intros; clear H2.
rewrite <- H4 in H1. clear H4 p0.
assert (paramOwner x = m /\ paramType x = t).
apply parameters_as_lists2. trivial.
destruct H2.
rewrite H0 in H4. rewrite <- H4 in H3.
tauto.
Qed.

Inductive ParamValueListTyping: Env -> Path -> ParamValueList -> Prop :=

  | ParamValueListTyping_nil:
    forall (G : Env) (p : Path),
      (ParamValueListTyping G p nil)

  | ParamValueListTyping_cons:
    forall (G : Env) (p : Path) (x : ParamSym) (v : Value) (xvs : ParamValueList) (T : type),
      (paramType x = T) ->
      (Gen_PathTyping G (Path_value v) (thisSubst_in_type T p)) ->
      (ParamValueListTyping G p xvs) ->
      (ParamValueListTyping G p ((x, v) :: xvs))
.
  
Lemma lemmaA40_aux_7:
  forall (G: Env) (xvs: ParamValueList) (xts: ParamTermList) (p: Path) (m: MethodSym),
    (ParamTermListTyping G p m xts) ->
    (paramValueList_to_ParamTermList xvs = xts) ->
    (ParamValueListTyping G p xvs)
.

Proof.
intros. generalize xvs xts H H0. clear xvs xts H H0.
induction xvs.
(* Nil *)
simpl. intros. apply ParamValueListTyping_nil.
(* Cons *)
simpl. intros. destruct a. rename p0 into x.
rewrite <- H0 in H. inversion_clear H.
eapply ParamValueListTyping_cons with (T := (paramType x)).
trivial.
rewrite H1. apply lemmaA40_aux_8. trivial.
eapply IHxvs. apply H3. trivial.
Qed.

Lemma lemmaA40_aux_11:
  forall (xvs: ParamValueList) (xts: ParamTermList),
    (paramValueList_to_ParamTermList xvs) = xts ->
    (params_of_paramValueList xvs) = (params_of_paramTermList xts)
.

Proof.
induction xvs.
simpl. intros. rewrite <- H. simpl. trivial.
simpl. destruct a. intros. rewrite <- H. simpl.
eapply f_equal with (f := (cons p)).
apply (IHxvs (paramValueList_to_ParamTermList xvs)). trivial.
Qed.

Lemma lemmaA40_aux_12:
  forall (m: MethodSym) (xs: (list ParamSym)),
    (IsComplete_params m xs) ->
    (List_sublist _ (params_of_paramTypeList (method_parameters m)) xs)
.

Proof.
intros.
unfold IsComplete_params in H.
unfold List_sublist.
intros.
unfold params_of_paramTypeList in H0.
pose (f := (fun xT : ParamSym * type => let (x, _) := xT in x)).
fold f in H0.
destruct (map_elim (ParamSym * type) (ParamSym) f (method_parameters m) x H0).
destruct x0.
unfold f in H1.
destruct H1. rewrite <- H2 in H1.
destruct (parameters_as_lists2 m x t H1).
apply H. trivial.
Qed.

Lemma lemmaA40_aux_13:
  forall (xTs: ParamTypeList) (p: Path),
    (params_of_paramTypeList (thisSubst_in_paramTypeList xTs p)) =
    (params_of_paramTypeList xTs)
.

Proof.
intros.
induction xTs.
simpl. trivial.
destruct a. simpl.
apply f_equal with (f := (cons p0)). trivial.
Qed.

Lemma lemmaA40_aux_14:
  forall (G: Env) (p: Path) (xvs: ParamValueList) (x: ParamSym) (v: Value) (T: type),
    (ParamValueListTyping G p xvs) ->
    (In (x, v) xvs) ->
    (paramType x = T) ->
    (Gen_PathTyping G (Path_value v) (thisSubst_in_type T p))
.

Proof.
intros. generalize xvs H H0. clear xvs H H0. 
induction xvs.
simpl. tauto.
destruct a. simpl. intros.
destruct H0. rewrite H0 in H.
inversion_clear H.
rewrite H1 in H2. rewrite H2. trivial.
apply IHxvs.
inversion_clear H. trivial. trivial.
Qed.

Lemma lemmaA40_aux_15:
  forall (xvs: ParamValueList) (m: MethodSym) (p: Path),
    (ParamValueListTyping rootEnv p xvs) ->
    (args_are_correct xvs (thisSubst_in_paramTypeList (method_parameters m) p))
.

Proof.
intros.
unfold args_are_correct.
intros.
unfold thisSubst_in_paramTypeList in H1.
pose (f :=
  (fun xT : ParamSym * type =>
    let (x, T) := xT in (x, thisSubst_in_type T p))
).
fold f in H1.
destruct (map_elim (ParamSym * type) (ParamSym * type) f (method_parameters m) (x, T) H1).
clear H1.
destruct x0. unfold f in H2.
destruct H2. injection H2; intros. clear H2.
rewrite H3. clear H3 T. rewrite <- H4 in H1. clear H4 p0.
destruct (parameters_as_lists2 m x t H1). clear H1.
eapply lemmaA40_aux_14. apply H. apply H0. trivial.
Qed.

(**********************************************************************)

Definition lemmaA40_prop1 (G: Env) (t: Term) (T: type) (H: (Typing G t T)): Prop :=
  G = rootEnv ->
  forall (u: Term),
    (Red_term t u) ->
    (Typing rootEnv u T)
.

Definition lemmaA40_prop2 (G: Env) (p: Path) (m: MethodSym) (xts: ParamTermList) (H: (ParamTermListTyping G p m xts)): Prop :=
  G = rootEnv ->
  forall (xts': ParamTermList),
    (Red_paramTermList xts xts') ->
    (ParamTermListTyping rootEnv p m xts')
.

Definition lemmaA40_prop3 (G: Env) (C: ClassSym) (fts: FieldTermList) (H: (FieldTermListTyping G C fts)): Prop :=
  G = rootEnv ->
  forall (fts': FieldTermList),
    (Red_fieldTermList fts fts') ->
    (FieldTermListTyping rootEnv C fts')
.

(** Subject-reduction **)
Lemma lemmaA40:
  forall (T: type) (t u: Term),
    (Typing rootEnv t T) ->
    (Red_term t u) ->
    (Typing rootEnv u T)
.

Proof.
cut (
  forall (G: Env) (t: Term) (T: type) (H: (Typing G t T)),
    (lemmaA40_prop1 G t T H)
).
intros.
unfold lemmaA40_prop1 in H.
eapply H. apply H0. trivial. trivial.
apply (Typing_prop
  lemmaA40_prop1
  lemmaA40_prop2
  lemmaA40_prop3
).

(* T-This *)
intros. red. intros.
rewrite H in e. discriminate e.

(* T-Param *)
intros. red. intros.
rewrite H in e.
injection e; intros.
rewrite <- H2 in i.
simpl in i. destruct i.

(* T-Var *)
intros. red. intros.
rewrite H in e.
injection e; intros.
destruct a.
rewrite <- H1 in H5.
inversion H5.

(* T-Select *)
intros. red. intros.
rewrite H in g. rewrite H in g0. clear H G.
rename T into T'. rename S into T. rename T' into S.
(* p = v , v = new B(fvs) *)
destruct (lemmaA7' p _ g). rename x into v.
destruct v. rename c into B.  rename f0 into fvs.
rewrite H in H0.
inversion_clear H0.
rename v into v'. pose (v := (Value_mk B fvs)).
assert (v = (Value_mk B fvs)). trivial.
assert (H': Gen_PathTyping rootEnv p (type_class C)). trivial.
rewrite H in g.
inversion_clear g.
inversion_clear H4.
(* |- v': U,  this: B |-_struct U <: S,  U closed  *)
destruct (lemmaA40_aux_1 rootEnv B fvs f v' S H3 H1 e0).
destruct H4. destruct H9. rename x into U.
(* this: B |- U <: S  *)
assert (Gen_Subtyping (classEnv B) U S).
apply lemmaA23_2. trivial.
(* |- v: B *)
assert (Gen_PathTyping rootEnv (Path_value v) (type_class B)).
rewrite H0.
apply Gen_P_Value. trivial. trivial.
apply IsComplete_intro. trivial. trivial. trivial.
apply Gen_S_Class.
(* U[v] = U *)
assert ((thisSubst_in_type U (Path_value v)) = U).
apply new_lemma1_2. unfold type_is_closed in H9. tauto.
(* |- U <: S[v] *)
assert (Gen_Subtyping rootEnv U (thisSubst_in_type S (Path_value v))).
rewrite <- H13.
unfold rootEnv.
rewrite <- (lemmaA40_aux_3 v).
rewrite <- (lemmaA40_aux_2 v).
eapply lemmaA39_this_2.
unfold classEnv in H11. apply H11. trivial.
(* |- U <: T *)
assert (Gen_Subtyping rootEnv U T).
rewrite H in g0.
rewrite <- H0 in g0.
eapply (Gen_S_Trans). apply H14. trivial.
unfold NoConstraint.applyTo. tauto.
(* |- v': T *)
eapply lemmaA1.
apply lemmaA40_aux_4.
apply H4. trivial.

(* Case T-Call *)
intros. red. intros.
rewrite H0 in g.
unfold lemmaA40_prop2 in H.
rewrite H0 in p0. rewrite H0 in g0. rewrite H0 in H. clear H0 G.
rename T into T'. rename S into T. rename T' into S.
(* p = v , v = new B(fvs) *)
destruct (lemmaA7' p _ g). rename x into v.
destruct v.
rename c into B. rename f into fvs.
pose (v := (Value_mk B fvs)).
assert (v = (Value_mk B fvs)). trivial.
rewrite <- H2 in H0.
rewrite H0 in H1.
inversion_clear H1.
(* R-Arg *)
eapply T_Call.
rewrite <- H0. apply g. trivial. apply e0.
rewrite <- (lemmaA40_aux_5 xts xts' H3). trivial.
rewrite <- H0. apply H. trivial. trivial.
rewrite <- H0. trivial.

(* R-Call *)
clear H.
rename C' into A. rename C0 into C'.
(* A, xTs |- t: S *)
assert (Typing (Env_mk (Owner_class A) (method_parameters m) nil) t S).
apply methodBody_of_type_methodType. trivial. trivial.
(* |- C' <: A *)
assert (Gen_Subtyping rootEnv (type_class C') (type_class A)).
apply lemmaA23_2. apply lemmaA22. trivial.
(* |- v: A *)
rewrite H0 in g. rewrite H3 in g. inversion_clear g.
assert (Gen_PathTyping rootEnv (Path_value v) (type_class A)).
rewrite H3.
apply Gen_P_Value. trivial.
trivial. trivial. trivial.
(* xTs[v] |- t[v]: S[v] *)
assert (Typing
  (Env_mk Owner_root (thisSubst_in_paramTypeList (method_parameters m) (Path_value v)) nil)
  (thisSubst_in_term t (Path_value v)) (thisSubst_in_type S (Path_value v))
).
rewrite <- (lemmaA40_aux_3 v).
apply lemmaA39_this_1 with (C := A). trivial. trivial.
(* this: C |- Ss WF *)
assert(
  forall (x: ParamSym) (U: type),
    (paramType x) = U ->
    (paramOwner x) = m ->
    (Kinding (classEnv C) U)
).
intros.
eapply param_types_are_well_formed. apply H13. apply H14. trivial.
(* fp(Ss) = empty *)
assert (
  forall (x: ParamSym) (U: type),
    (paramType x) = U ->
    (paramOwner x) = m ->
    (fp_in_type U) = empty_paramset
).
intros.
assert (Kinding (classEnv C) U). eapply H13. apply H14. trivial.
unfold empty_paramset.
apply empty_is_empty_2 with (A := ParamSym). apply ParamSym_eq.
intros.
fold (paramset_In x0 (fp_in_type U)). intro.
unfold classEnv in H16.
assert (In x0 nil).
Goal (
  (paramTypeList_to_paramList nil) = nil
).
simpl. trivial.
Save R1.
rewrite <- R1.
eapply lemmaA3_3. apply H16. trivial.
simpl in H18. tauto.
(* fp(Ss[v]) = empty *)
assert (
  forall (x: ParamSym) (U: type),
    (paramType x) = U ->
    (paramOwner x) = m ->
    (fp_in_type (thisSubst_in_type U (Path_value v))) = empty_paramset
).
intros.
rewrite <- (H14 x U H15 H16).
apply new_lemma2_2.
simpl. trivial.
clear H13 H14.
(* Env(root; xs: S[v]; epsilon) is well-formed *)
assert (context_is_well_formed
  (Env_mk Owner_root (thisSubst_in_paramTypeList (method_parameters m) (Path_value v)) nil)
).
apply context_is_well_formed_intro.
apply forall_is_listForall.
intros.
destruct x. rename p1 into x. rename t0 into Uv.
pose (U := (paramType x)).
assert ((paramType x) = U). trivial.
destruct (lemmaA40_aux_6 _ _ _ _ _ H13 H14).
rewrite H17.
eapply H15. apply H14. trivial.
unfold local_types_well_formed.
simpl. tauto. clear H15.
(* |- t[v][xvs]: S[v][xvs] *)
assert (Typing rootEnv
  (map_paramSubst_in_term (thisSubst_in_term t (Path_value v)) xvs)
  (map_paramSubst_in_type (thisSubst_in_type S (Path_value v)) xvs)
).
unfold rootEnv.
replace (nil (A := type)) with (map_paramSubst_in_typeList nil xvs).
eapply lemmaA39_param_1 with
  (xTs := (thisSubst_in_paramTypeList (method_parameters m) (Path_value v)))
  (Us := (nil (A := type)))
  (G := (Env_mk Owner_root
    (thisSubst_in_paramTypeList (method_parameters m) (Path_value v))
    nil))
. trivial. trivial. trivial.
unfold args_conform_to_params.
split.
rewrite lemmaA40_aux_13.
apply lemmaA40_aux_12.
rewrite (lemmaA40_aux_11 xvs xts). trivial. trivial.
apply lemmaA40_aux_15.
eapply lemmaA40_aux_7.
rewrite <- H0. apply p0. trivial.
simpl. trivial.
(* this: C |- S WF *)
clear H12 H13.
assert (Kinding (classEnv C) S).
eapply methodTypes_are_well_formed. apply e. trivial.
(* fp(S) = emptyset *)
assert ((fp_in_type S) = empty_paramset).
unfold empty_paramset.
apply empty_is_empty_2 with (A := ParamSym). apply ParamSym_eq.
intros.
fold (paramset_In x (fp_in_type S)).
intro.
cut (In x (paramTypeList_to_paramList nil)). simpl. tauto.
eapply lemmaA3_3. unfold classEnv in H12. apply H12. trivial.
(* S[v][xvs] = S[v] *)
assert (
  (map_paramSubst_in_type (thisSubst_in_type S (Path_value v)) xvs) =
  (thisSubst_in_type S (Path_value v))
).
cut ((map_paramSubst_in_path (Path_value v) xvs) = (Path_value v)). intros.
pattern (Path_value v) at 2. rewrite <- H15.
apply lemmaA9_2.
intros. unfold param_in_type. rewrite H13. simpl. tauto.
rewrite polySubst_in_pathValue. trivial.
(* |- t[v][xvs]: T *)
rewrite H15 in H14. clear H15.
eapply lemmaA1. apply H14.
rewrite <- H0. trivial.

(* Case T-New *)
intros. red. intros.
inversion H1.
clear H4 fts0. clear H2 C0. clear H3 H1 u.
apply T_New.
assert ((fields_of_fieldTermList fts) = (fields_of_fieldTermList fts')).
apply lemmaA40_aux_16. trivial.
rewrite <- H1. trivial.
eapply H. trivial. trivial.
assert ((fields_of_fieldTermList fts) = (fields_of_fieldTermList fts')).
apply lemmaA40_aux_16. trivial.
rewrite <- H1. trivial.
rewrite <- H0. trivial.

(* Case T-Let *)
(* T-Let (resume) *)
intros. red. intros.
inversion H2.
(* R-Local *)
eapply T_Let. rewrite H1 in e. apply e.
apply H. trivial. trivial. apply t1. trivial.
rewrite <- H1. trivial.
(* R-Let *)
clear u0 H2 H4. clear T0 H3. clear t2 H5. clear u1 H6.
rewrite H1 in e.
clear H0.
unfold rootEnv in e. injection e; intros. clear e.
rewrite <- H0 in t1. simpl in t1. clear H0.
rewrite <- H2 in t1. clear H2.
rewrite <- H3 in t1. clear O H3.
clear H.
(* |- v: T *)
assert (Gen_PathTyping rootEnv (Path_value v) T).
apply lemmaA40_aux_8. rewrite H7. rewrite <- H1. trivial.
clear t0 H7.
(* fv(U) = emptyset (delayed) *)
cut ((fv_in_type U) = empty_varset). intros.
apply lemmaA1 with (T := U).
rewrite <- (lemmaA37_2 U 0 v).
eapply lemmaA39_var_1_corollary. apply t1. trivial.
trivial.
rewrite <- (lemmaA31_2 U). rewrite <- H1. trivial.
trivial.
(* fv(U) = emptyset *)
unfold empty_varset.
apply empty_is_empty_2 with (A := VarSym). apply VarSym_eqdec.
intros. intro.
destruct x.
(* x = 0 *)
apply n.
apply fv_and_var_in_type2. trivial.
fold (varset_In (Datatypes.S x) (fv_in_type U)) in H0.
assert ((Datatypes.S x) < (length (T :: nil))).
eapply lemmaA4_1. apply t1.
unfold varset_In. unfold varset_union.
apply set_union_intro2. trivial.
simpl in H2.
assert (x < 0). omega.
unfold lt in H3.
apply (le_Sn_O x). trivial.

(* Case T-Nil (param/term) *)
intros. red. intros.
inversion H0.

(* Case T-Cons (param/term) *)
intros. red. intros.
inversion H2.
(* R-Head *)
clear x0 t1 xts0 xts' H2 H3 H5 H6 H4.
eapply ParamTermListTyping_cons. apply e.
apply H. trivial. trivial. rewrite <- H1. trivial.
(* R-Tail *)
clear x0 t1 xts0 H3 H4 H6.
eapply ParamTermListTyping_cons. apply e.
rewrite <- H1. trivial.
apply H0. trivial. trivial.

(* Case T-Nil (field/term) *)
intros. red. intros.
inversion H0.

(* Case T-Cons (field/term) *)
intros. red. intros.
inversion H2.
(* R-Head *)
clear f1 t2 fts0 fts' H2 H3 H5 H6 H4.
eapply FieldTermListTyping_cons.
apply H. trivial. trivial. trivial. apply e. trivial.
rewrite <- H1. trivial.
(* R-Tail *)
clear f1 t2 fts0 H3 H4 H6.
eapply FieldTermListTyping_cons.
rewrite <- H1. apply t0. trivial. apply e. trivial.
apply H0. trivial. trivial.
Qed.

End SetProgram.
