(* Author: Vincent Cremet *)
(* Date: July 2006 *) 
(* File: WellFoundation.v *)
(* Description: Lemmas about well-founded relations *)

(** * Some Properties of Well-foundation *)

Require Import List.

(** Accessible => no direct cycles **)
Lemma well_foundation1:
  forall (A: Set) (R: A -> A -> Prop) (x: A),
    (Acc R x) ->
    forall (y: A),
      (R x y) ->
      (R y x) -> False.

Proof.
intros A R x H.
induction H.
intros.
apply (H0 y H2 x H2 H1).
Qed.

(** Well-founded => no direct cycles **)
Lemma well_foundation2:
  forall (A: Set) (R: A -> A -> Prop),
    (well_founded R) ->
    forall (x y: A),
      (R x y) ->
      (R y x) -> False.

Proof.
intros.
unfold well_founded in H.
assert (Acc R x). apply H. clear H.
apply (well_foundation1 A R x H2 y H0 H1).
Qed.

(** Function extension of a binary relation **)
Definition fun_extension (A B: Set) (f: A -> B) (R: B -> B -> Prop) (x y: A): Prop :=
  (R (f x) (f y))
.

(** accessible interpretation **)
Lemma well_foundation3:
  forall (A B: Set) (f: A -> B) (R: B -> B -> Prop),
    forall (x: B),
      (Acc R x) ->
      forall (x': A),
        (x = (f x')) ->
        (Acc (fun_extension A B f R) x')
.

Proof.
intros A B f R x H.
induction H.
intros.
apply Acc_intro.
intros.
red in H2.
rewrite -> H1 in H.
rewrite -> H1 in H0. clear H1.
assert (Acc R (f y)).
apply H. trivial.
apply (H0 (f y)). trivial.
trivial.
Qed.

(** Well-founded interpretation **)
Lemma well_foundation4:
  forall (A B: Set) (f: A -> B) (R: B -> B -> Prop),
    (well_founded R) -> 
    (well_founded (fun_extension A B f R))
.

Proof.
intros.
red in H.
red.
intro.
apply (well_foundation3 A B f R (f a)).
trivial. trivial.
Qed.
